@@ -1,7 +1,8 @@
 #!/usr/bin/env python
 
 #--------------------------------------------------------------------------------------
-## VulaSHAKA (Simultaneous Neutronic, Fuel Performance, Heat And Kinetics Analysis)
+## pythonFlu - Python wrapping for OpenFOAM C++ API
+## Copyright (C) 2010- Alexey Petrov
 ## Copyright (C) 2009-2010 Pebble Bed Modular Reactor (Pty) Limited (PBMR)
 ## 
 ## This program is free software: you can redistribute it and/or modify
@@ -17,7 +18,7 @@
 ## You should have received a copy of the GNU General Public License
 ## along with this program.  If not, see <http://www.gnu.org/licenses/>.
 ## 
-## See https://vulashaka.svn.sourceforge.net/svnroot/vulashaka
+## See http://sourceforge.net/projects/pythonflu
 ##
 ## Author : Alexey PETROV
 ##
@@ -27,14 +28,14 @@
 def createFluidMeshes( rp, runTime ) :
     
     from Foam.finiteVolume import fvMesh, PtrList_fvMesh
-    fluidRegions = PtrList_fvMesh( rp.fluidRegionNames.size() )
+    fluidRegions = PtrList_fvMesh( rp.fluidRegionNames().size() )
     
     from Foam.OpenFOAM import word, fileName, IOobject
-    for index in range( rp.fluidRegionNames.size() ) :
+    for index in range( rp.fluidRegionNames().size() ) :
         from Foam.OpenFOAM import ext_Info, nl
-        ext_Info()<< "Create fluid mesh for region " << rp.fluidRegionNames[ index ] \
+        ext_Info()<< "Create fluid mesh for region " << rp.fluidRegionNames()[ index ] \
                   << " for time = " << runTime.timeName() << nl << nl
-        mesh = fvMesh( IOobject( rp.fluidRegionNames[ index ],
+        mesh = fvMesh( IOobject( rp.fluidRegionNames()[ index ],
                                  fileName( runTime.timeName() ),
                                  runTime,
                                  IOobject.MUST_READ ) )
@@ -60,17 +61,23 @@ def createFluidFields( fluidRegions, runTime ) :
 
     DpDtFluid = PtrList_volScalarField( fluidRegions.size() )
      
-    from Foam.OpenFOAM import PtrList_UniformDimensionedVectorField
-    gFluid = PtrList_UniformDimensionedVectorField( fluidRegions.size() )
+    from Foam.OpenFOAM import PtrList_uniformDimensionedVectorField
+    gFluid = PtrList_uniformDimensionedVectorField( fluidRegions.size() )
  
     from Foam.compressible import PtrList_compressible_turbulenceModel
     turbulence = PtrList_compressible_turbulenceModel( fluidRegions.size() ) 
     
-    from Foam.thermophysicalModels import PtrList_basicPsiThermo
-    thermoFluid = PtrList_basicPsiThermo( fluidRegions.size() ) 
+    from Foam.thermophysicalModels import PtrList_basicRhoThermo
+    thermoFluid = PtrList_basicRhoThermo( fluidRegions.size() ) 
     
-    from Foam.OpenFOAM import List_scalar
-    initialMassFluid = List_scalar( fluidRegions.size() )
+    p_rghFluid = PtrList_volScalarField( fluidRegions.size() )
+    
+    ghFluid = PtrList_volScalarField( fluidRegions.size() )
+    
+    ghfFluid = PtrList_surfaceScalarField(fluidRegions.size())
+    
+    from Foam.OpenFOAM import scalarList
+    initialMassFluid = scalarList( fluidRegions.size() )
     
     #Populate fluid field pointer lists
 
@@ -81,9 +88,9 @@ def createFluidFields( fluidRegions, runTime ) :
 
         ext_Info()<< "    Adding to thermoFluid\n" << nl
         
-        from Foam.thermophysicalModels import autoPtr_basicPsiThermo, basicPsiThermo
+        from Foam.thermophysicalModels import autoPtr_basicRhoThermo, basicRhoThermo
  
-        thermo= basicPsiThermo.New( fluidRegions[ index ] )
+        thermo= basicRhoThermo.New( fluidRegions[ index ] )
         thermoFluid.ext_set( index, thermo )
         
         ext_Info()<< "    Adding to rhoFluid\n" << nl
@@ -125,8 +132,8 @@ def createFluidFields( fluidRegions, runTime ) :
                                                      linearInterpolate( rhoFluid[ index ] * UFluid[ index ] ) & fluidRegions[ index ].Sf() ) )
         
         ext_Info()<< "    Adding to gFluid\n" << nl
-        from Foam.OpenFOAM import UniformDimensionedVectorField
-        gFluid.ext_set( index, UniformDimensionedVectorField( IOobject( word( "g" ),
+        from Foam.OpenFOAM import uniformDimensionedVectorField
+        gFluid.ext_set( index, uniformDimensionedVectorField( IOobject( word( "g" ),
                                                                         fileName( runTime.constant() ),
                                                                         fluidRegions[ index ],
                                                                         IOobject.MUST_READ,
@@ -138,17 +145,31 @@ def createFluidFields( fluidRegions, runTime ) :
                                                                      UFluid[ index ],
                                                                      phiFluid[ index ],
                                                                      thermoFluid[ index ] ) )
+        ext_Info() << "    Adding to ghFluid\n" << nl
+        ghFluid.ext_set( index, volScalarField( word( "gh" ) , gFluid[ index ] & fluidRegions[ index ].C() ) )
+
+        ext_Info() << "    Adding to ghfFluid\n" << nl
+        ghfFluid.ext_set( index, surfaceScalarField( word( "ghf" ), gFluid[ index ] & fluidRegions[ index ].Cf() ) )
+
+        p_rghFluid.ext_set( index, volScalarField( IOobject( word( "p_rgh" ),
+                                                         fileName( runTime.timeName() ),
+                                                         fluidRegions[ index ],
+                                                         IOobject.MUST_READ,
+                                                         IOobject.AUTO_WRITE ),
+                                               fluidRegions[ index ] ) )
+        # Force p_rgh to be consistent with p
+        p_rghFluid[ index ].ext_assign( thermoFluid[ index ].p() - rhoFluid[ index ] * ghFluid[ index ] )
+        
+        from Foam import fvc
+        initialMassFluid[ index ] = fvc.domainIntegrate( rhoFluid[ index ] ).value() 
         
         ext_Info()<< "    Adding to DpDtFluid\n" << nl
-        from Foam import fvc
         DpDtFluid.ext_set( index, volScalarField( word( "DpDt" ), fvc.DDt( surfaceScalarField( word( "phiU" ), 
                                                                                                phiFluid[ index ] / fvc.interpolate( rhoFluid[ index ] ) ),
                                                                            thermoFluid[ index ].p() ) ) )
 
-        initialMassFluid[ index ] = fvc.domainIntegrate( rhoFluid[ index ] ).value()              
-        pass
     
-    return thermoFluid, rhoFluid, KFluid, UFluid, phiFluid, gFluid, turbulence, DpDtFluid, initialMassFluid
+    return thermoFluid, rhoFluid, KFluid, UFluid, phiFluid, gFluid, turbulence, DpDtFluid, initialMassFluid, ghFluid, ghfFluid, p_rghFluid
         
 
 #-----------------------------------------------------------------------------------------------------------------------
@@ -199,7 +220,7 @@ def readFluidMultiRegionPIMPLEControls( mesh ) :
 
 #--------------------------------------------------------------------------------------------------------------------------
 def setRegionFluidFields( i, fluidRegions, thermoFluid, rhoFluid, KFluid, UFluid, \
-                          phiFluid, gFluid, turbulence, DpDtFluid, initialMassFluid ):
+                          phiFluid, turbulence, DpDtFluid, initialMassFluid, ghFluid, ghfFluid, p_rghFluid ):
     mesh = fluidRegions[ i ]
 
     thermo = thermoFluid[ i ]
@@ -207,19 +228,22 @@ def setRegionFluidFields( i, fluidRegions, thermoFluid, rhoFluid, KFluid, UFluid
     K = KFluid[ i ]
     U = UFluid[ i ]
     phi = phiFluid[ i ]
-    g = gFluid[ i ]
-
+    
     turb = turbulence[ i ]
     DpDt = DpDtFluid[ i ]
 
     p = thermo.p()
     psi = thermo.psi()
     h = thermo.h()
-
+    
+    p_rgh = p_rghFluid[ i ]
+    gh = ghFluid[ i ]
+    ghf = ghfFluid[ i ]
+    
     from Foam.OpenFOAM import dimensionedScalar, word, dimMass
-    massIni = dimensionedScalar( word( "massIni" ), dimMass, initialMassFluid[ i ] )
+    initialMass = dimensionedScalar( word( "massIni" ), dimMass, initialMassFluid[ i ] )
 
-    return mesh, thermo, rho, K, U, phi, g, turb, DpDt, p, psi, h, massIni
+    return mesh, thermo, rho, K, U, phi, turb, DpDt, p, psi, h, initialMass, p_rgh, gh, ghf
 
 
 #--------------------------------------------------------------------------------------------------------------------------
@@ -231,8 +255,8 @@ def storeOldFluidFields( p, rho ):
 
 #--------------------------------------------------------------------------------------------------------------------------
 def initContinuityErrs( size ):
-    from Foam.finiteVolume import List_scalar
-    cumulativeContErr = List_scalar( size, 0.0 )
+    from Foam.OpenFOAM import scalarList
+    cumulativeContErr = scalarList( size, 0.0 )
     
     return cumulativeContErr
 
@@ -259,7 +283,7 @@ def compressibleContinuityErrors( i, mesh, rho, thermo, cumulativeContErr ) :
 
 
 #--------------------------------------------------------------------------------------------------------------------------
-def fun_UEqn( rho, U, phi, g, p, turb, mesh, momentumPredictor ) :
+def fun_UEqn( rho, U, phi, ghf, p_rgh, turb, mesh, momentumPredictor ) :
     # Solve the Momentum equation
     from Foam import fvm
     UEqn = fvm.ddt( rho, U ) + fvm.div( phi, U ) + turb.divDevRhoReff( U )
@@ -269,7 +293,7 @@ def fun_UEqn( rho, U, phi, g, p, turb, mesh, momentumPredictor ) :
     if momentumPredictor :
         from Foam import fvc
         from Foam.finiteVolume import solve
-        solve( UEqn == fvc.reconstruct( fvc.interpolate( rho ) * ( g & mesh.Sf() )  - fvc.snGrad( p ) * mesh.magSf() ) )
+        solve( UEqn == fvc.reconstruct( ( -ghf * fvc.snGrad( rho ) - fvc.snGrad( p_rgh ) )*mesh.magSf() ) )
         pass
     
     return UEqn
@@ -300,10 +324,16 @@ def fun_hEqn( rho, h, phi, turb, DpDt, thermo, mesh, oCorr, nOuterCorr ) :
 
 
 #---------------------------------------------------------------------------------------------------------    
-def fun_pEqn( i, mesh, p, g, rho, turb, thermo, thermoFluid, K, UEqn, U, phi, psi, DpDt, massIni, \
+def fun_pEqn( i, mesh, p, rho, turb, thermo, thermoFluid, K, UEqn, U, phi, psi, DpDt, initialMass, p_rgh, gh, ghf, \
               nNonOrthCorr, oCorr, nOuterCorr, corr, nCorr, cumulativeContErr ) :
     
-    closedVolume = p.needReference()
+    closedVolume = p_rgh.needReference()
+    
+    from Foam import fvc
+    compressibility = fvc.domainIntegrate( psi )
+    
+    from Foam.OpenFOAM import SMALL
+    compressible = ( compressibility.value() > SMALL )
 
     rho.ext_assign( thermo.rho() )
     
@@ -315,31 +345,39 @@ def fun_pEqn( i, mesh, p, g, rho, turb, thermo, thermoFluid, K, UEqn, U, phi, ps
     rhorUAf = surfaceScalarField( word( "(rho*(1|A(U)))" ) , fvc.interpolate( rho * rUA ) )
 
     U.ext_assign( rUA * UEqn.H() ) 
-    
+
     from Foam import fvc
 
     phiU = ( fvc.interpolate( rho ) *
                  (  ( fvc.interpolate( U ) & mesh.Sf() ) +
                       fvc.ddtPhiCorr( rUA, rho, U, phi ) ) )
-    phi.ext_assign( phiU + fvc.interpolate( rho ) * ( g & mesh.Sf() ) * rhorUAf )
-    
+    phi.ext_assign( phiU - rhorUAf * ghf * fvc.snGrad( rho ) * mesh.magSf() )
     from Foam import fvm
+    
+    from Foam.finiteVolume import correction
+    p_rghDDtEqn = fvc.ddt( rho ) + psi * correction( fvm.ddt( p_rgh ) ) + fvc.div( phi )
+    
+    # Thermodynamic density needs to be updated by psi*d(p) after the
+    # pressure solution - done in 2 parts. Part 1:
+    thermo.rho().ext_assign( thermo.rho() - psi * p_rgh )
+
     for nonOrth in range ( nNonOrthCorr + 1 ):
-        pEqn = ( fvm.ddt( psi, p) + fvc.div( phi ) - fvm.laplacian( rhorUAf, p ) )
-        if oCorr == ( nOuterCorr - 1 ) and corr == ( nCorr - 1 ) and nonOrth == nNonOrthCorr :
-            pEqn.solve( mesh.solver( word( str( p.name() ) +  "Final" ) ) )
-        else:
-            pEqn.solve( mesh.solver( p.name() ) )
-            pass
+        p_rghEqn = p_rghDDtEqn - fvm.laplacian( rhorUAf, p_rgh )
+        p_rghEqn.solve( mesh.solver( p_rgh.select( ( oCorr == nOuterCorr-1 and corr == nCorr-1 and nonOrth == nNonOrthCorr ) ) ) )
         
         if nonOrth == nNonOrthCorr :
-            phi.ext_assign( phi + pEqn.flux() )
-            pass
+           phi.ext_assign( phi + p_rghEqn.flux() )
+           pass
         pass
     
+    # Second part of thermodynamic density update
+    thermo.rho().ext_assign( thermo.rho() + psi * p_rgh )
+    
     # Correct velocity field
     U.ext_assign( U + rUA * fvc.reconstruct( ( phi - phiU ) / rhorUAf ) )
     U.correctBoundaryConditions()
+    
+    p.ext_assign( p_rgh + rho * gh )
 
     #Update pressure substantive derivative
     DpDt.ext_assign( fvc.DDt( surfaceScalarField( word( "phiU" ), phi / fvc.interpolate( rho ) ), p ) )
@@ -353,9 +391,10 @@ def fun_pEqn( i, mesh, p, g, rho, turb, thermo, thermoFluid, K, UEqn, U, phi, ps
     
     # For closed-volume cases adjust the pressure and density levels
     # to obey overall mass continuity
-    if closedVolume :
-       p.ext_assign( p + ( massIni - fvc.domainIntegrate( psi * p ) ) / fvc.domainIntegrate( psi ) )
+    if closedVolume and compressible:
+       p.ext_assign( p + ( initialMass - fvc.domainIntegrate( thermo.rho() ) ) / compressibility )
        rho.ext_assign( thermo.rho() )
+       p_rgh.ext_assign( p - rho * gh )
        pass
     #Update thermal conductivity
     K.ext_assign( thermoFluid[ i ].Cp() * turb.alphaEff() )
@@ -364,7 +403,7 @@ def fun_pEqn( i, mesh, p, g, rho, turb, thermo, thermoFluid, K, UEqn, U, phi, ps
 
 
 #--------------------------------------------------------------------------------------------------------------------------
-def solveFluid( i, mesh, thermo, thermoFluid, rho, K, U, phi, g, h, turb, DpDt, p, psi, massIni,\
+def solveFluid( i, mesh, thermo, thermoFluid, rho, K, U, phi, h, turb, DpDt, p, psi, initialMass, p_rgh, gh, ghf,\
                 oCorr, nCorr, nOuterCorr, nNonOrthCorr, momentumPredictor,cumulativeContErr ) :
     
     if oCorr == 0 :
@@ -372,12 +411,12 @@ def solveFluid( i, mesh, thermo, thermoFluid, rho, K, U, phi, g, h, turb, DpDt,
         rhoEqn( rho, phi )
         pass
     
-    UEqn = fun_UEqn( rho, U, phi, g, p, turb, mesh, momentumPredictor )
+    UEqn = fun_UEqn( rho, U, phi, ghf, p_rgh, turb, mesh, momentumPredictor )
     hEqn = fun_hEqn( rho, h, phi, turb, DpDt, thermo, mesh, oCorr, nOuterCorr )
     
     # --- PISO loop
     for corr in range( nCorr ):
-        cumulativeContErr =  fun_pEqn( i, mesh, p, g, rho, turb, thermo, thermoFluid, K, UEqn, U, phi, psi, DpDt, massIni,
+        cumulativeContErr =  fun_pEqn( i, mesh, p, rho, turb, thermo, thermoFluid, K, UEqn, U, phi, psi, DpDt, initialMass, p_rgh, gh, ghf,
                                        nNonOrthCorr, oCorr, nOuterCorr, corr, nCorr, cumulativeContErr )
         pass
     
